#include "PullMoveSet.hh"
#include "NeighborVector.hh"

namespace biu
{
	PullMoveSet::PullMoveDecoder::PullMoveDecoder(const LatticeModel* lattice) 
	{
		LatticeNeighborhood neighbors = lattice->getNeighborhood();
		
		// save all neighbor vectors for reference
		for (LatticeNeighborhood::const_iterator it = neighbors.begin();
				it != neighbors.end(); it++)
		{
			neighDirections.push_back(*it);
		}
		directionNumber = neighDirections.size();
		
		// set up stdPulls
		// structure: direction i->(x) : pull number : (c, l)
		
		// for every direction between two neighbors
		for (LatticeNeighborhood::const_iterator direction = neighbors.begin();
				direction != neighbors.end(); direction++)
		{
			std::vector< std::vector< IntPoint > > stdMoves;
			// for each neighbor of i
			for (LatticeNeighborhood::const_iterator nbi = neighbors.begin();
					nbi != neighbors.end(); nbi++)
			{
				// for each neighbor of (x)
				if (*nbi != *direction)
				for (LatticeNeighborhood::const_iterator nb = neighbors.begin();
				nb != neighbors.end(); nb++)
				{
					// if neighbor of i and neighbor of (x) are neighbored
					// save new c,l pair
					if ((*direction + *nb) != IntPoint(0,0,0) &&
							lattice->areNeighbored( *nbi, (*direction + *nb) ))
					{
						std::vector< IntPoint > clvec;
						clvec.push_back( *nbi );
						clvec.push_back( (*direction + *nb) );
						stdMoves.push_back( clvec );
					}
				}
			}
			stdPulls.push_back( stdMoves );
		}
		stdPullNumber = stdPulls.at(0).size();
		
		// set up endPulls
		for (LatticeNeighborhood::const_iterator nb1 = neighbors.begin();
			nb1 != neighbors.end(); nb1++)
		{
			for (LatticeNeighborhood::const_iterator nb2 = neighbors.begin();
				nb2 != neighbors.end(); nb2++)
			{
				if ((*nb2 + *nb1) != IntPoint(0,0,0))
				{
					IPointVec pullVec;
					pullVec.push_back(*nb1);
					pullVec.push_back(*nb1 + *nb2);
					endPulls.push_back(pullVec);
				}
			}
		}
		endPullNumber = (directionNumber-1)*(directionNumber-1);
	}
	
	void 
	PullMoveSet::PullMoveDecoder::lookupMove(const size_t& todoSize,
			size_t& moveIndex, bool& pullFront, 
			bool& stdPull, size_t& movePosition) const
	{
		// *********************************
		// code table:
		// *********************************
		// legal moves: 
		// 		[0, endPullNumber*2 + (proteinLength-1)*stdPullNumber*2 - 1]
		// end pulls @ position 0:
		//		[endPullNumber, 2*endPullNumber - 1]
		// end pulls @ last position:
		//		[0, endPullNumber - 1]
		// standart pulls:
		//		[2*endPullNumber, 
		//			endPullNumber*2 + (proteinLength-1)*stdPullNumber*2 - 1]
		
		if (moveIndex < endPullNumber) 
		{
			// end pull at end position
			pullFront=true;
			stdPull=false;
			movePosition=todoSize-1;
			return;
		}
		else if (moveIndex < endPullNumber*2)
		{
			// end pull at position 0
			pullFront=false;
			stdPull=false;
			movePosition=0;
			moveIndex-=endPullNumber;
			return;
		}
		else
		{
			// do std pulls
			stdPull = true;
			moveIndex -= endPullNumber*2;
			if (moveIndex < stdPullNumber)
			{
				// pull first element & pull front
				pullFront = true;
				movePosition = 0;
				return;
			}
			else if (moveIndex < stdPullNumber*2)
			{
				// pull last element & pull back
				moveIndex -= stdPullNumber;
				pullFront = false;
				movePosition = todoSize-1;
				return;
			}
			else
			{
				// do standard pull @ element in [1, todoSize-1]
				moveIndex -= stdPullNumber*2;
				movePosition = 1 + moveIndex/(2*stdPullNumber);
				pullFront = (moveIndex % (2*stdPullNumber)) < stdPullNumber ?
						true : false;
				moveIndex = moveIndex % stdPullNumber;
				return;
			}
		}
	}

	IPointVec 
	PullMoveSet::PullMoveDecoder::lookupEndMove(const IntPoint& rootpoint,
			const size_t& moveIndex) const
	{
		IPointVec movePositions;
		movePositions.push_back(endPulls.at(moveIndex).at(0) + rootpoint);
		movePositions.push_back(endPulls.at(moveIndex).at(1) + rootpoint);
		return movePositions;
	}
	
	IPointVec
	PullMoveSet::PullMoveDecoder::lookupCL(const IntPoint& pullPoint,
			const IntPoint& fixedPoint, 
			const size_t& moveIndex) const
	{
		const IntPoint moveDirection = fixedPoint - pullPoint;
		
		for (int i = 0;	i < directionNumber; i++)
		{
			// check where to look in stdPulls
			if (neighDirections.at(i) == moveDirection) 
			{
				return stdPulls.at(i).at(moveIndex);
			}
		}
		
		return IPointVec();
	}
	
	size_t
	PullMoveSet::PullMoveDecoder::getMoveNumber(const size_t proteinLength) const 
	{
		return	endPullNumber*2 +		// end pulls
				(proteinLength-1)*stdPullNumber*2;					// std pulls
	}
	
	PullMoveSet::PullMoveSet(const LatticeModel* lattice)
	 :	LatticeMoveSet(lattice)
	{
	 	undoRec.lastChangedObject = NULL;
		decoder = new PullMoveSet::PullMoveDecoder(lattice);
	} 

	PullMoveSet::~PullMoveSet()
	{
		delete decoder;
	}

	LatticeProtein*
	PullMoveSet::applyMove(const LatticeProtein* const todo, 
			size_t position, size_t moveIndex)  {

		// apply the pivot move inplace to a copy of todo
		return applyMoveInPlace(new LatticeProtein(*todo), position,
				 moveIndex);
	}
	
	LatticeProtein* 
	PullMoveSet::applyMoveInPlace(LatticeProtein* todo, 
			size_t position, size_t moveIndex)  
	{
		
		undoRec.hasChanged = false;
		const size_t todoSize = todo->getSequence().size();

		assertbiu(moveIndex >=0 && moveIndex < getMoveNumber(todo),
			"The pull move moveIndex has to be in [0, number of available moves)." 
			<< getMoveNumber(todo));
		assertbiu(position >=0 && position < todoSize,
			"The pull move position has to be in [0, structure length).");

		bool pullFront;
		bool stdPull;
		decoder->lookupMove(todoSize, moveIndex, pullFront, stdPull, position);
		
		// RelativeInt objects are used to be able to compute both pull 
		// directions the same code
		const RelativeInt relPos(position, pullFront);
		const RelativeInt relLastElement((pullFront ? 0 : todoSize-1), pullFront);
		
		// do end pull
		if (!stdPull) 
		{
			IPointVec pullPoints = 
				decoder->lookupEndMove(todo->getPoints()->at(position), 
										moveIndex);
			
			// move legal? -> pull positions not occupied
			for (IPointVec::const_iterator it = todo->points->begin(); 
					it != todo->points->end(); it++)
			{
				// if position occupied, leave
				if ( (*it == pullPoints.at(0)) || (*it == pullPoints.at(0)) ) 
					return todo;
			}
			
			// save undo information
			undoRec.hasChanged = true;
			undoRec.lastChangedObject = todo;
			undoRec.lostPos0 = todo->getPoints()->at(relLastElement.getValue());
			undoRec.lostPos1 = todo->getPoints()->at(relLastElement+1);
			undoRec.position = position;
			undoRec.lastPullPosition = relLastElement.getValue();
			undoRec.pullFront = pullFront;
			
			// pull tail
			for (RelativeInt i(relLastElement.getValue(), pullFront);
					i <= relPos -2; i++)
			{
				todo->getPoints()->at(i.getValue()) = todo->getPoints()->at(i+2);
			}
			
			todo->points->at(relPos-1) = pullPoints.at(0);
			todo->points->at(relPos.getValue()) = pullPoints.at(1);

			// compute new relative movestring
			MoveSequence newMoves = lattice->pointsToRelMoves(*(todo->points));
			todo->relMoves->swap(newMoves);
			
			return todo;
		}

		// look up pullPoint & fixedPoint
		const IntPoint pullPoint = todo->getPoints()->at(position);
		const IntPoint fixedPoint = todo->getPoints()->at(relPos + 1);
				
		// look up c & l
		const IPointVec clvec = 
			decoder->lookupCL(pullPoint, fixedPoint, moveIndex);
		const IntPoint c = clvec[0] + pullPoint;
		const IntPoint l = clvec[1] + pullPoint;
		
		// check if C and L are valid positions
		// if C position occupied by element in pull direction
		// or if moving first element while pulling front
		// or if moving last element while pulling back
		if ( (pullFront && position == 0) 
				|| (!pullFront && position == todoSize-1) 
				|| (c==todo->points->at(relPos - 1)) )
		// just check for L position
		{
			for (IPointVec::const_iterator it = todo->points->begin(); 
				it != todo->points->end(); it++)
			{
	 			// if L position occupied, leave
				if ((*it) == l) return todo;
			}
		}
		// check both C and L positions
		else
		{
			for (IPointVec::const_iterator it = todo->points->begin(); 
			it != todo->points->end(); it++)
			{
				// if C or L position occupied, leave
				if (((*it) == l) || ((*it) == c)) return todo;
			}
		}

		// compute how much needs to be moved
		int lastPullPosition=-1;					// last element to be moved
		bool lastPullPositionFound = false;

		// if moving first element while pulling front
		// or if moving last element while pulling back
		// or if moving second element
		if (relPos <= relLastElement + 1)
		{
			lastPullPosition = relLastElement.getValue();
			lastPullPositionFound = true;
		}
		// if moving anything above second element
		else
		{
			// if c and position-2 are neighbored
			if (lattice->areNeighbored(c, todo->points->at(relPos-2)))
			{
				// just change position and position-1
				lastPullPosition = relPos-1;
				lastPullPositionFound = true;
			}
		}
		// if moving anything above third element
		if (relPos >= relLastElement + 3
				&& !lastPullPositionFound)
		{
			// check beyond position-2
			for (RelativeInt i(position, pullFront); relLastElement <= i-3; i--)
			{
				if (lattice->areNeighbored(
						todo->points->at(i.getValue()), 
						todo->points->at(i-3)))
				{
					// pull until position i+1
					lastPullPosition = i-2;
					break;
				}
			}
		}
		if (!lastPullPositionFound) lastPullPosition = relLastElement.getValue();
		assertbiu(lastPullPosition!=-1, 
				"Unhandled case in PullMoveSet move calculation.");
		
		// save values for undo:
		undoRec.lastChangedObject = todo;
		undoRec.hasChanged = true;
		undoRec.pullFront = pullFront;
		undoRec.position = position;
		undoRec.lastPullPosition = lastPullPosition;
		
		// move
		const RelativeInt relLastPullPosition(lastPullPosition, pullFront);
		if (relLastPullPosition <= relPos-2)
		{
			// save for undo
			undoRec.lostPos0 = todo->points->at(relLastPullPosition.getValue());
			undoRec.lostPos1 = todo->points->at(relLastPullPosition+1);
			// pull tail if existent
			for (RelativeInt i(lastPullPosition, pullFront); i<=relPos-2; i++)
			{
				todo->points->at(i.getValue()) = todo->points->at(i+2);
			}
			todo->points->at(relPos-1) = c;
			todo->points->at(relPos.getValue()) = l;
		}
		else if (lastPullPosition == relPos-1)
		{
			// save for undo
			undoRec.lostPos0 = todo->points->at(relPos-1);
			todo->points->at(relPos-1) = c;
			undoRec.lostPos1 = todo->points->at(relPos.getValue());
			todo->points->at(relPos.getValue()) = l;
		}
		else if (lastPullPosition == relPos.getValue())
		{
			// save for undo
			undoRec.lostPos1 = todo->points->at(relPos.getValue());
			todo->points->at(relPos.getValue()) = l;
		}
		
		// compute new relative movestring
		MoveSequence newMoves = lattice->pointsToRelMoves(*(todo->points));
		todo->relMoves->swap(newMoves);
		
		return todo;
	}

	unsigned int
	PullMoveSet::getMoveNumber() const {
		// number of pull moves fixed for CUB and standard moves -> 8
		return 8;
	}
	
	size_t 
	PullMoveSet::getMoveNumber(const LatticeProtein * const lp) const
	{
		return decoder->getMoveNumber(lp->getSequence().size());
	}

	LatticeProtein* 
	PullMoveSet::undoLastMove(LatticeProtein* toUndo) 
	{
		const RelativeInt relPos(undoRec.position, undoRec.pullFront);
		const RelativeInt relLastPullPos
								(undoRec.lastPullPosition, undoRec.pullFront);
		
		if (toUndo == undoRec.lastChangedObject && undoRec.hasChanged)
		{
			if (undoRec.lastPullPosition == relPos.getValue())
			{
				toUndo->points->at(relPos.getValue()) = undoRec.lostPos1;
			}
			if (undoRec.lastPullPosition == relPos-1)
			{
				toUndo->points->at(relPos.getValue()) = undoRec.lostPos1;
				toUndo->points->at(relPos-1) = undoRec.lostPos0;
			}
			if (relLastPullPos <= relPos-2)
			{
				// undo pulling whole tail
				for (RelativeInt i(undoRec.position, undoRec.pullFront);
						i>=relLastPullPos+2; i--)
				{
					toUndo->points->at(i.getValue()) = toUndo->points->at(i-2);
				}
				toUndo->points->at(relLastPullPos.getValue()) = undoRec.lostPos0;
				toUndo->points->at(relLastPullPos + 1) = undoRec.lostPos1;
			}
			
			// compute new relative movestring
			MoveSequence newMoves = lattice->pointsToRelMoves(*(toUndo->points));
			toUndo->relMoves->swap(newMoves);
		}
		
		return toUndo;
	}
}
